class CommitmentLevel:
    Processed: "CommitmentLevel"
    Confirmed: "CommitmentLevel"
    Finalized: "CommitmentLevel"
    def __int__(self) -> int: ...
    def __str__(self) -> str: ...
    def __repr__(self) -> str: ...
    def __eq__(self, o: object) -> bool: ...
    @staticmethod
    def from_string(s: str) -> "CommitmentLevel": ...
    @staticmethod
    def default() -> "CommitmentLevel": ...

class CommitmentConfig:
    def __init__(self, commitment: "CommitmentLevel") -> None: ...
    @property
    def commitment(self) -> "CommitmentLevel": ...
    @staticmethod
    def from_string(s: str) -> "CommitmentConfig": ...
    @staticmethod
    def processed() -> "CommitmentConfig": ...
    @staticmethod
    def confirmed() -> "CommitmentConfig": ...
    @staticmethod
    def finalized() -> "CommitmentConfig": ...
    @staticmethod
    def default() -> "CommitmentConfig": ...
    def is_finalized(self) -> bool: ...
    def is_confirmed(self) -> bool: ...
    def is_at_least_confirmed(self) -> bool: ...
